#include "kscreenmanager.h"

#include <KF5/KScreen/kscreen/output.h>
#include <KF5/KScreen/kscreen/edid.h>
#include <KF5/KScreen/kscreen/mode.h>
#include <KF5/KScreen/kscreen/config.h>
#include <KF5/KScreen/kscreen/getconfigoperation.h>
#include <KF5/KScreen/kscreen/configmonitor.h>
#include <KF5/KScreen/kscreen/setconfigoperation.h>
#include <KF5/KScreen/kscreen/edid.h>
#include <KF5/KScreen/kscreen/types.h>
using namespace KScreen;

#include <QDebug>
#include <QDBusInterface>
#include <QDBusConnection>
#include <QProcess>
#include <unistd.h>
#include <QStandardPaths>
#include <QProcess>
#include <QException>

KScreenManager::KScreenManager(QObject *parent) : QObject(parent)
{
    // 使用usd设置显示参数
    m_strSaveParamPath = QStandardPaths::findExecutable("save-param");
}

KScreenManager::~KScreenManager()
{

}

void KScreenManager::initManager()
{
    requestBackend();
//    connect(new KScreen::GetConfigOperation(), &KScreen::GetConfigOperation::finished,
//                this, &KScreenManager::onKScreenConfigfinished);
    auto op = new KScreen::GetConfigOperation();
    KScreen::ConfigPtr configPtr = nullptr;
    if (!op->exec()) {
        qWarning("Failed to retrieve backend: %s", qPrintable(op->errorString()));
        configPtr = KScreen::ConfigPtr();
    } else {
        configPtr = op->config();
    }
    onConfigUpdate(configPtr);
}

void KScreenManager::onKScreenConfigfinished(ConfigOperation *operation)
{
    onConfigUpdate(qobject_cast<KScreen::GetConfigOperation *>(operation)->config());
}

void KScreenManager::onConfigUpdate(ConfigPtr ptrConfig)
{
    if (!ptrConfig || !ptrConfig->isValid())
        return;
    if (m_ptrConfig) {
        KScreen::ConfigMonitor::instance()->removeConfig(m_ptrConfig);
        for (const KScreen::OutputPtr &output : m_ptrConfig->outputs()) {
            output->disconnect(this);
        }
        m_ptrConfig->disconnect(this);
    }
    m_ptrConfig = ptrConfig;

    KScreen::ConfigMonitor::instance()->addConfig(m_ptrConfig);
    qDebug()<<"ScreenConfig:"<< m_ptrConfig->connectedOutputs();

    connect(m_ptrConfig.data(), &KScreen::Config::outputAdded,
            this, &KScreenManager::onOutputAdded);
    connect(m_ptrConfig.data(), &KScreen::Config::outputRemoved,
            this, &KScreenManager::onOutputRemoved);

    connect(m_ptrConfig.data(), &KScreen::Config::primaryOutputChanged,
            this, &KScreenManager::onPrimaryChanged);
    if (m_isFirstLoad) {
        // 更新显示输出信息
        for (const KScreen::OutputPtr &output : m_ptrConfig->outputs()) {
            if (output->isConnected()) {
                onOutputConnected(output);
            }
            onOutputAdded(output);
        }
        adjustPerfectConfig();
        m_isFirstLoad = false;
    }
}

void KScreenManager::onOutputAdded(const OutputPtr& output)
{
    connect(output.data(), &KScreen::Output::isConnectedChanged,
            this, &KScreenManager::onOutputConnectChanged);
    connect(output.data(), &KScreen::Output::isEnabledChanged,
            this, &KScreenManager::onOutputEnableChanged);
    qDebug()<<"onOutputAdded -----";
}

void KScreenManager::onOutputRemoved(int nOutput)
{
    KScreen::OutputPtr output = m_ptrConfig->output(nOutput);
    if (!output.isNull()) {
        output->disconnect(this);
    }
    qDebug()<<"onOutputRemoved -----";
}

void KScreenManager::onPrimaryChanged(const OutputPtr& output)
{

}

void KScreenManager::onOutputConnectChanged()
{
    const KScreen::OutputPtr output(qobject_cast<KScreen::Output *>(sender()), [](void *){});
    if (!output) {
        return;
    }

    if (output->isConnected()) {
        onOutputConnected(output);
    } else {
        onOutputDisconnected(output->id());
    }
}

void KScreenManager::onOutputConnected(const KScreen::OutputPtr& output)
{
    qDebug()<<"onOutputConnected:"<<output->name()<<","<<output->id();
    if (!m_isFirstLoad) {
        adjustPerfectConfig();
    }
}

void KScreenManager::onOutputDisconnected(int nOutput)
{
    qDebug()<<"onOutputDisconnected:"<<nOutput;
    if (!m_isFirstLoad) {
        adjustPerfectConfig();
    }
}

void KScreenManager::onOutputEnableChanged()
{

}

void KScreenManager::setModesType(ScreensMode mode)
{
    m_curMode = mode;
}

void KScreenManager::adjustPerfectConfig()
{
    int nConnectedCount = 0;
    qDebug()<<"Adjust ScreenConfig:"<< m_ptrConfig->connectedOutputs();
    if (!m_ptrConfig)
        return;
    for (const KScreen::OutputPtr &output : m_ptrConfig->outputs()) {
        if (output->isConnected()) {
            nConnectedCount ++;
        }
    }
    if (nConnectedCount < 1) {
        return ;
    }
    switch (m_curMode) {
    case SCREENS_MODE_ORI:
    {
        for (const KScreen::OutputPtr &output : m_ptrConfig->outputs()) {
            if (output->isConnected()) {
                if (output->isPrimary()) {
                    output->setFollowPreferredMode(true);
                    output->setCurrentModeId(output->preferredModeId());
                    double lfScale = chkScreenScale(output->preferredMode()->size().width(),
                                                    output->preferredMode()->size().height());
                    output->setScale(lfScale);
                    output->setEnabled(true);
                } else {
                    output->setEnabled(false);
                }
            }
        }
    }
        break;
    case SCREENS_MODE_CLONE:
    {
        //查找最佳克隆分辨率
        OutputList outputList = m_ptrConfig->outputs();
        ModeList sameList = outputList[0]->modes();
        for (const OutputPtr &output : outputList) {
            if (output->isConnected()) {
                ModeList::iterator itMode = sameList.begin();
                for ( ; itMode != sameList.end(); ) {
                    if (!output->modes().contains(itMode.key())) {
                        itMode = sameList.erase(itMode);
                    } else {
                        itMode ++;
                    }
                }
            }
        }
        if (sameList.size() > 0) {
            QList<int> clones;
            KScreen::OutputPtr baseOutput = nullptr;
            for (const KScreen::OutputPtr &output : m_ptrConfig->outputs()) {
                if (output->isConnected()) {
                    if (!baseOutput) {
                        baseOutput = output;
                    }
                    output->setCurrentModeId(sameList.begin().key());
                    double lfScale = chkScreenScale(sameList.begin().value()->size().width(),
                                                    sameList.begin().value()->size().height());
                    output->setScale(lfScale);
                    output->setEnabled(true);
                    clones << output->id();
                    output->setPos(QPoint(0, 0));
                }
            }
            if (baseOutput) {
                baseOutput->setClones(clones);
            }
        }
    }
        break;
    case SCREENS_MODE_ONLY_OUT:
    {
        if (nConnectedCount < 2) {
            break;
        }
        OutputList outputList = m_ptrConfig->outputs();
        int nConnectedIndex = 0;
        int nXPos = 0;
        OutputList::iterator itOutput = outputList.begin();
        for ( ; itOutput != outputList.end(); itOutput++) {
            if (itOutput.value()->isConnected()) {
                if (nConnectedIndex == 0) {
                    itOutput.value()->setEnabled(false);
                } else {
                    if (nConnectedIndex == 1) {
                        itOutput.value()->setPrimary(true);
                    }
                    itOutput.value()->setFollowPreferredMode(true);
                    itOutput.value()->setCurrentModeId(itOutput.value()->preferredModeId());
                    double lfScale = chkScreenScale(itOutput.value()->preferredMode()->size().width(),
                                                    itOutput.value()->preferredMode()->size().height());
                    itOutput.value()->setScale(lfScale);
                    itOutput.value()->setEnabled(true);
                    itOutput.value()->setPos(QPoint(nXPos, 0));
                    itOutput.value()->setClones(QList<int>());
                    lfScale = lfScale == 0 ? 1 : lfScale;
                    nXPos += (double)(itOutput.value()->preferredMode()->size().width()) / lfScale;
                }
                nConnectedIndex ++;
            }
        }
    }
        break;
    case SCREENS_MODE_EXTEND:
    default:
    {
        OutputList outputList = m_ptrConfig->outputs();
        OutputList::iterator itOutput = outputList.begin();
        OutputPtr primaryOutput = nullptr;
        for ( ; itOutput != outputList.end(); itOutput++) {
            if (itOutput.value()->isConnected()) {
                if (itOutput.value()->isPrimary()) {
                    primaryOutput = itOutput.value();
                    break;
                }
            }
        }
        if (primaryOutput) {
            int nXPos = 0;
            primaryOutput->setFollowPreferredMode(true);
            primaryOutput->setPos(QPoint(nXPos, 0));
            primaryOutput->setClones(QList<int>());
            qDebug()<<"PrimaryInfo:"<<primaryOutput->preferredModeId()<<primaryOutput->preferredModes()<<
                      primaryOutput->preferredMode()->size().width()<<primaryOutput->preferredMode()->size().height();
            primaryOutput->setCurrentModeId(primaryOutput->preferredModeId());
            double lfScale = chkScreenScale(primaryOutput->preferredMode()->size().width(),
                                            primaryOutput->preferredMode()->size().height());
            primaryOutput->setScale(lfScale);
            primaryOutput->setEnabled(true);
            lfScale = lfScale == 0 ? 1 : lfScale;
            nXPos += (double)(primaryOutput->preferredMode()->size().width()) / lfScale;

            itOutput = outputList.begin();
            for ( ; itOutput != outputList.end(); itOutput++) {
                if (itOutput.value()->isConnected() && primaryOutput != itOutput.value()) {
                    itOutput.value()->setPrimary(false);
                    itOutput.value()->setFollowPreferredMode(true);
                    itOutput.value()->setPos(QPoint(nXPos, 0));
                    itOutput.value()->setClones(QList<int>());
                    qDebug()<<"OtherInfo:"<<itOutput.value()->preferredModeId()<<itOutput.value()->preferredModes()<<
                              itOutput.value()->preferredMode()->size().width()<<itOutput.value()->preferredMode()->size().height();
                    itOutput.value()->setCurrentModeId(itOutput.value()->preferredModeId());
                    double lfScale = chkScreenScale(itOutput.value()->preferredMode()->size().width(),
                                                    itOutput.value()->preferredMode()->size().height());
                    itOutput.value()->setScale(lfScale);
                    itOutput.value()->setEnabled(true);
                    lfScale = lfScale == 0 ? 1 : lfScale;
                    nXPos += (double)(itOutput.value()->preferredMode()->size().width()) / lfScale;
                }
            }
        } else {
            int nXPos = 0;
            itOutput = outputList.begin();
            for ( ; itOutput != outputList.end(); itOutput++) {
                if (itOutput.value()->isConnected()) {
                    if (nXPos == 0) {
                        itOutput.value()->setPrimary(true);
                    }
                    itOutput.value()->setFollowPreferredMode(true);
                    itOutput.value()->setPos(QPoint(nXPos, 0));
                    itOutput.value()->setClones(QList<int>());
                    itOutput.value()->setCurrentModeId(itOutput.value()->preferredModeId());
                    double lfScale = chkScreenScale(itOutput.value()->preferredMode()->size().width(),
                                                    itOutput.value()->preferredMode()->size().height());
                    itOutput.value()->setScale(lfScale);
                    itOutput.value()->setEnabled(true);
                    lfScale = lfScale == 0 ? 1 : lfScale;
                    nXPos += (double)(itOutput.value()->preferredMode()->size().width()) / lfScale;
                }
            }
        }
    }
        break;
    }
    applyConfig();
    Q_EMIT screenCountChanged(nConnectedCount);
}

void KScreenManager::applyConfig()
{
    const KScreen::ConfigPtr &config = m_ptrConfig;
    qDebug() << "NewScreenConfig:" << config->connectedOutputs();

    if (!KScreen::Config::canBeApplied(config)) {
        qWarning()<<"Sorry, your configuration could not be applied!";
        return;
    }

    /* Store the current config, apply settings */
    auto *op = new KScreen::SetConfigOperation(config);

    /* Block until the operation is completed, otherwise KCMShell will terminate
     * before we get to execute the Operation */
    op->exec();
    sync();
}

void KScreenManager::requestBackend()
{
    QDBusInterface screenIft("org.kde.KScreen",
                             "/",
                             "org.kde.KScreen",
                             QDBusConnection::sessionBus());
    if (!screenIft.isValid()) {
        QProcess process;
        process.start("uname -m");
        process.waitForFinished();
        QString output = process.readAll();
        output = output.simplified();

        QString command = "/usr/lib/" + output + "-linux-gnu"
                          +"/libexec/kf5/kscreen_backend_launcher";
        QProcess::startDetached(command);
    }
}

double KScreenManager::chkScreenScale(int nWidth, int nHeight)
{
    double  mScaling = 1.0;
    if (nWidth <= 1920 && nHeight <=1080){
        mScaling = 1.0;
    }
    else if (nWidth > 1920 && nHeight > 1080 && nWidth <= 2560 && nHeight <=1500) {
        mScaling = 1.5;
    }
    else if (nWidth > 2560 && nHeight > 1440) {
        mScaling = 2.0;
    }
    return mScaling;
}

void KScreenManager::resumeDisplayParam(QString strUserName)
{
    if (!m_strSaveParamPath.isEmpty() && !strUserName.isEmpty()) {
        QProcess subProcess;
        try {
            QString usdSaveParam = m_strSaveParamPath + " -u " +strUserName;
            qDebug() << "setCurUserName usdSaveParam: " << usdSaveParam;
            subProcess.start(usdSaveParam);
            subProcess.waitForFinished();
        } catch(const QException &e) {
            qWarning() << e.what();
        }
    }
}
