#ifndef KSCREENMANAGER_H
#define KSCREENMANAGER_H

#include <QObject>

#include <KF5/KScreen/kscreen/types.h>

enum ScreensMode{
    SCREENS_MODE_ORI        = 1,
    SCREENS_MODE_CLONE      = 2,
    SCREENS_MODE_EXTEND     = 3,
    SCREENS_MODE_ONLY_OUT   = 4
};

namespace KScreen {
class ConfigOperation;
}

class KScreenManager : public QObject
{
    Q_OBJECT
public:
    explicit KScreenManager(QObject *parent = nullptr);
    virtual ~KScreenManager();

    void initManager();
    void adjustPerfectConfig();
    void setModesType(ScreensMode mode);
    void resumeDisplayParam(QString strUserName);

private Q_SLOTS:
    void onKScreenConfigfinished(KScreen::ConfigOperation *operation);
    void onOutputAdded(const KScreen::OutputPtr& output);
    void onOutputRemoved(int nOutput);
    void onPrimaryChanged(const KScreen::OutputPtr& output);
    void onOutputConnectChanged();
    void onOutputEnableChanged();
    void onOutputConnected(const KScreen::OutputPtr& output);
    void onOutputDisconnected(int nOutput);

signals:
    void screenCountChanged(int nCount);

private:
    void requestBackend();
    void onConfigUpdate(KScreen::ConfigPtr ptrConfig);
    void applyConfig();
    double chkScreenScale(int nWidth, int nHeight);

private:
    KScreen::ConfigPtr m_ptrConfig = nullptr;
    ScreensMode m_curMode = SCREENS_MODE_EXTEND;
    bool m_isFirstLoad = true;
    QString m_strSaveParamPath = "";
};

#endif // KSCREENMANAGER_H
